# frozen_string_literal: true

class ComplexSubject < ActiveRecord::Base
  include ComplexHelper
  include HyraxHelper
  include S3FileHandleable

  attr_accessor :is_imported

  has_many :complex_sessions, foreign_key: 'parent_source_identifier', primary_key: 'source_identifier', dependent: :destroy
  has_many :complex_modalities, through: :complex_sessions

  attribute :source_identifier, :string, default: -> { SecureRandom.uuid }

  validates :source_identifier, presence: true, uniqueness: true
  validates :parent_source_identifier, presence: true
  validates :s3_folder_name, presence: { message: ->(object, data) { I18n.t('rdms.errors.complex_subject.title.blank', value: object.s3_folder_name) } }
  validates_length_of :s3_folder_name, within: 3..63, too_short: ->(object, data) { I18n.t('rdms.errors.complex_subject.title.length', value: object.s3_folder_name) }, too_long: ->(object, data) { I18n.t('rdms.errors.complex_subject.title.length', value: object.s3_folder_name) }
  validates :s3_folder_name, uniqueness: { scope: :parent_source_identifier,  message: ->(object, data) { I18n.t('rdms.errors.complex_subject.title.uniqueness', value: object.s3_folder_name) } }, unless: -> { self.is_imported }

  before_validation :set_s3_folder_name
  after_save :save_metadata_as_json_in_s3, :update_crc_dataset_date_modified, :relocate_files_on_folder_name_change, :save_meta_json_in_s3
  after_commit :fix_duplication_of_title, on: :create
  after_commit :clear_s3_and_all_associated_objects, on: [:destroy]

  scope :sort_by_title, ->(source_identifier) {
    where(parent_source_identifier: source_identifier)
    .sort_by { |item| item.subject_title.downcase.gsub(/[^a-zA-Z0-9]/, '0').split(/(\d+)/).map { |s| s.match(/\d+/) ? s.to_i : s } }
  }

  def crc_dataset
    CrcDataset.where(source: parent_source_identifier).first
  end

  def file_sets
    FileSet.where(for_complex_identifier: source_identifier)
  end

  def file_sets_fast_load_metadata
    FileSetFastLoadMetadata.sort_and_filter_by_title_and_source(source_identifier)
  end

  def s3_bucket_and_object_prifix
    [crc_dataset.id, "#{s3_folder_name}/"]
  end

  private

  def set_s3_folder_name
    self.s3_folder_name = sanitize_title(subject_title)
  end

  def save_metadata_as_json_in_s3
    s3 = S3StorageService.new
    s3.init_client
    bucket_name = s3.sanitise_name(crc_dataset.id)
    s3.create_bucket(bucket_name) unless s3.bucket_exists?(bucket_name)
    s3.add_content(bucket_name, "#{s3_folder_name}/system_metadata.json", to_json)
    s3.add_content(bucket_name, "#{s3_folder_name}/metadata.json", filter_metadata_json_file(self))
  end

  def update_crc_dataset_date_modified
    crc_dataset.update(date_modified: DateTime.now)
  end

  def fix_duplication_of_title
    return true unless self.is_imported

    suffix = 1
    new_title = self.subject_title

    while ComplexSubject.where(subject_title: new_title, parent_source_identifier: crc_dataset.source.first).where.not(id: self.id).present?
      new_title = "#{self.subject_title}_#{suffix}"
      suffix += 1
    end

    self.subject_title = new_title
    self.save
  end
end
