# frozen_string_literal: true
require 'fileutils'
require 'csv'

class S3StorageService
  # Reference: https://docs.aws.amazon.com/sdk-for-ruby/v3/developer-guide/
  # https://docs.ceph.com/en/latest/radosgw/s3/ruby/
  # https://docs.aws.amazon.com/sdk-for-ruby/v3/developer-guide/s3-example-create-buckets.html
  attr_reader :s3_client

  def init_client
    @s3_client = Aws::S3::Client.new
  end

  def sanitise_name(bucket_name)
    if ENV.fetch('S3_BUCKET_PREFIX', '')
      "#{ENV['S3_BUCKET_PREFIX']}-#{bucket_name}".downcase.gsub(/[^0-9a-z.-]/, '')
    else
      bucket_name.downcase.gsub(/[^0-9a-z.-]/, '')
    end
  end

  def list_buckets
    buckets = []
    @s3_client.list_buckets.buckets.each do |bucket|
      buckets << {
        name: bucket.name,
        creation_date: bucket.creation_date,
        size: bucket.size
      }
    end
    buckets
  end

  def system_metadata_object(bucket_name)
    @s3_client.get_object(bucket: bucket_name, key: "system_metadata.json")
  end

  def bucket_exists?(bucket_name)
    @s3_client.list_buckets.buckets.each do |bucket|
      return true if bucket.name == bucket_name
    end
    false
  end

  def object_exists?(bucket_name, object_key)
    return true if @s3_client.get_object(bucket: bucket_name, key: object_key)
  rescue Aws::S3::Errors::NoSuchKey => e
    false
  end

  def list_objects(bucket_name, prefix = nil, max_keys = 1000)
    # To get the first max_keys objects in a bucket
    contents = []
    total_size = 0
    objects = @s3_client.list_objects_v2(
      bucket: bucket_name,
      max_keys: max_keys,
      prefix: prefix
    ).contents
    objects.each do |object|
      contents << {
        key: object.key,
        size: object.size,
        last_modified_date: object.last_modified
      }
      total_size += object.size
    end
    return contents, total_size
  end

  def list_all_objects(bucket_name, prefix = nil)
    list_of_objects = []
    total_size = 0
    resp = @s3_client.list_objects_v2(
      bucket: bucket_name,
      max_keys: 1000,
      prefix: prefix
    )
    loop do
      objects = resp.contents
      objects.each do |object|
        list_of_objects << {
          key: object.key,
          size: object.size,
          last_modified_date: object.last_modified
        }
        total_size += object.size
      end
      break unless resp.next_page?

      resp = resp.next_page
    end
    return list_of_objects, total_size
  end

  def get_content(bucket_name, object_key, local_file_path)
    @s3_client.get_object(
      bucket: bucket_name,
      key: object_key,
      response_target: local_file_path
    )
  end

  def create_bucket(bucket_name)
    response = @s3_client.create_bucket(bucket: bucket_name)
    return true if response.location == "/#{bucket_name}"

    false
  end

  def add_content(bucket_name, object_key, object_content)
    response = s3_client.put_object(
      bucket: bucket_name,
      key: object_key,
      body: object_content
      # content_type: 'text/plain'
    )
    return true if response.etag

    false
  end

  def add_multipart_content(_bucket_name, _object_key, _file_path)
    # ToDo
    false
  end

  def delete_object(bucket_name, object_key)
    response = @s3_client.delete_objects(
      bucket: bucket_name,
      delete: {
        objects: [
          {
            key: object_key
          }
        ]
      }
    )
    return true if response.deleted.count == 1

    false
  end

  def delete_objects(bucket_name, object_keys)
    delete_keys = []
    object_keys.each do |object_key|
      delete_keys << { key: object_key }
    end

    response = @s3_client.delete_objects(
      bucket: bucket_name,
      delete: {
        objects: delete_keys
      }
    )
    return true if response.deleted.count == object_keys.count

    false
  end

  def delete_all_objects(bucket_name)
    Aws::S3::Bucket.new(bucket_name, client: @s3_client).clear!
  end

  def delete_bucket(bucket_name)
    response = @s3_client.delete_bucket(bucket: bucket_name)
  end

  def move_object(source_object_key, target_bucket_name, target_object_key, metadata = {}, source_bucket_name = nil)
    file_upload_bucket = source_bucket_name || sanitise_name(ENV['S3_FILE_UPLOAD_BUCKET'])
    source_bucket = Aws::S3::Resource
                             .new(region: ENV['S3_REGION'])
                             .bucket(file_upload_bucket)

    source_object = source_bucket.object(source_object_key)

    source_object.move_to(bucket: target_bucket_name, key: target_object_key, metadata: metadata)
  rescue Aws::Errors::ServiceError => e
    puts "Couldn't copy #{source_object.key} to #{target_object_key}. Here's why: #{e.message}"
  end

  def download_bucket(bucket, download_path, list_of_objects: nil)
    s3_bucket = sanitise_name(bucket)
    return false unless bucket_exists?(s3_bucket)
    list_of_objects, _total_size = list_all_objects(s3_bucket) unless list_of_objects
    list_of_objects.each do |object|
      path = File.join(download_path, object[:key])
      dirname = File.dirname(path)
      FileUtils.mkdir_p(dirname) unless File.directory?(dirname)
      unless object[:key].include?("system_metadata.json")
        get_content(s3_bucket, object[:key], path)
      end
    end
    download_path
  end

  def prepare_shell_file(bucket, shell_file, list_of_objects: nil)
    s3_bucket = sanitise_name(bucket)
    return false unless bucket_exists?(s3_bucket)

    list_of_objects, _total_size = list_all_objects(s3_bucket) unless list_of_objects

    File.open(shell_file, "w") do |file|
      file.write("#!/bin/bash\n\n")

      file = generate_curl_commands(file, s3_bucket, bucket, list_of_objects)

      file.write(")\n\n")

      file = shell_file_download_progress_bar(file)
    end

    shell_file
  end

  def prepare_search_result_shell_file(buckets, shell_file_path)
    download_dir = File.dirname(shell_file_path)
    FileUtils.mkdir_p(download_dir) unless  File.directory?(download_dir)
    File.open(shell_file_path, "w") do |file|
      file.write("#!/bin/bash\n")
      file.write("curl_commands=(\n")

      buckets.each do |each_bucket|
        bucket_with_prefix = Array(each_bucket)
        bucket = bucket_with_prefix[0]
        prefix = bucket_with_prefix.size > 1 ? bucket_with_prefix[1] : nil
        s3_bucket = sanitise_name(bucket)
        next unless bucket_exists?(s3_bucket)
        list_of_objects, _total_size = list_all_objects(s3_bucket, prefix)
        file = generate_curl_commands(file, s3_bucket, bucket, list_of_objects)
      end

      file.write(")\n\n")

      file = shell_file_download_progress_bar(file)
    end
  end

  def cleanup(buckets_to_keep=['fowi-rdms-testbucket'], run=false)
    buckets_to_delete = []
    list_buckets.each do |b|
      if b[:name].start_with?("#{ENV['S3_BUCKET_PREFIX']}-".downcase.gsub(/[^0-9a-z.-]/, '')) and
        not buckets_to_keep.include?(b[:name])
        buckets_to_delete << b[:name]
      end
    end
    return buckets_to_delete unless run
    buckets_to_delete.each do |b|
      puts "Deleting #{b}"
      delete_all_objects(b)
      delete_bucket(b)
    end
  end

  def presigned_url_for_download(bucket_name, file_key, expiry = ENV.fetch('S3_EXPIRY', 1))
    bucket = Aws::S3::Resource.new(region: ENV['S3_REGION']).bucket(bucket_name)
    url_options = {
      response_content_disposition: "attachment; filename=\"#{file_key}\""
    }

    # S3_EXPIRY is in days. expiry.days.to_i converts it to seconds
    url_options[:expires_in] = expiry.to_i.days.to_i if expiry.present?

    object = bucket.object(file_key)
    object.exists? ? object.presigned_url(:get, url_options).to_s : nil
  end


  def copy_object(source_bucket_name, target_bucket_name, source_object_key, target_object_key)
    source_bucket = Aws::S3::Bucket.new(source_bucket_name)
    source_object = source_bucket.object(source_object_key)
    target_bucket_name = sanitise_name(target_bucket_name)

    source_object.copy_to(bucket: target_bucket_name, key: target_object_key)

    puts "#{target_object_key} copied successfully!"  # Confirm successful copy
    presigned_url_for_download(target_bucket_name, target_object_key)
  rescue Aws::S3::Errors::NoSuchKey
    puts "Warning: Object #{target_object_key} not found in source bucket."
  end

  def generate_curl_commands(file, s3_bucket, folder_name, list_of_objects)
    list_of_objects.each_with_object([]) do |object, commands|
      unless object[:key].include?("system_metadata.json")
        key = File.join(folder_name, object[:key])
        file.write("    \"'#{key}' '#{presigned_url_for_download(s3_bucket, object[:key])}'\"\n")
      end
    end

    file
  end

  def shell_file_download_progress_bar(file)
    file.write("errors=()\n")
    file.write("downloaded_files=0\n")
    file.write("total_files=${#curl_commands[@]}\n\n")

    file.write("# Function to display the progress bar\n")
    file.write("show_progress() {\n")
    file.write("    local current=$1\n")
    file.write("    local total=$2\n")
    file.write("    local width=50\n")
    file.write("    local percent=$((current * 100 / total))\n")
    file.write("    local filled=$((percent * width / 100))\n")
    file.write("    local empty=$((width - filled))\n")
    file.write("    printf \"[\"\n")
    file.write("    printf \"%0.s#\" $(seq 1 $filled)\n")
    file.write("    printf \"%0.s \" $(seq 1 $empty)\n")
    file.write("    printf \"] %d%% (%d/%d)\\r\" $percent $current $total\n")
    file.write("}\n\n")

    file.write("# Function to handle each curl command\n")
    file.write("download_file() {\n")
    file.write("    local output_file=\"$1\"\n")
    file.write("    local url=\"$2\"\n")
    file.write("    local http_code=$(curl -f -s -w '%{http_code}' --create-dirs -o \"$output_file\" \"$url\")\n")
    file.write("    if [ $http_code -eq 200 ]; then\n")
    file.write("        downloaded_files=$((downloaded_files + 1))\n")
    file.write("        show_progress $downloaded_files $total_files\n")
    file.write("    else\n")
    file.write("        errors+=(\"$http_code | $url\")\n")
    file.write("        if [ -f \"$output_file\" ]; then\n")
    file.write("          rm -f \"$output_file\"\n")
    file.write("        fi\n")
    file.write("    fi\n")
    file.write("}\n\n")

    file.write("# Iterate over the curl commands and download files\n")
    file.write("for cmd in \"${curl_commands[@]}\"; do\n")
    file.write("    eval download_file $cmd\n")
    file.write("done\n\n")

    file.write("# Print final progress\n")
    file.write("show_progress $downloaded_files $total_files\n")
    file.write("echo\n\n")

    file.write("# Print errors if any\n")
    file.write("echo \"The download of all files is complete.\"\n")
    file.write("if [ ${#errors[@]} -ne 0 ]; then\n")
    file.write("    echo \"The following files could not be downloaded:\"\n")
    file.write("    for error in \"${errors[@]}\"; do\n")
    file.write("        echo \"$error\"\n")
    file.write("    done\n")
    file.write("else\n")
    file.write("    echo \"There were no errors in downloading the files.\"\n")
    file.write("fi\n")

    file
  end
end
